open import Relation.Binary.PropositionalEquality
open import Data.Unit using (⊤ ; tt)
open import Data.Product renaming (_×_ to _∧_ ; proj₁ to fst ; proj₂ to snd)
open import Data.Sum renaming (_⊎_ to _∨_ ; inj₁ to left ; inj₂ to right)
open import Data.Nat using (ℕ ; zero ; suc ; _+_)
open import Data.Product renaming (proj₁ to fst ; proj₂ to snd)
open import Data.List hiding (length ; head)

data _≤_ : ℕ → ℕ → Set where
  lez : {n : ℕ}                 → zero  ≤ n
  lesuc : {m n : ℕ} (m≤n : m ≤ n) → suc m ≤ suc n

{- 1 : Order on natural numbers -}

≤-pred : {m n : ℕ} → (suc m ≤ suc n) → m ≤ n
≤-pred (lesuc e) = e

≤-suc : {m n : ℕ} → m ≤ n → suc m ≤ suc n
≤-suc e = lesuc e

≤s : (n : ℕ) → n ≤ suc n
≤s zero = lez
≤s (suc n) = ≤-suc (≤s n)


≤-refl : {n : ℕ} → n ≤ n
≤-refl {zero} = lez
≤-refl {suc n} = lesuc ≤-refl

≤-trans : {m n p : ℕ} → (m ≤ n) → (n ≤ p) → (m ≤ p)
≤-trans lez e2 = lez
≤-trans (lesuc e1) (lesuc e2) = lesuc (≤-trans e1 e2)

_≤?_ : (m n : ℕ) → (m ≤ n) ∨ (n ≤ m)
zero ≤? n = left lez
suc m ≤? zero = right lez
suc m ≤? suc n with (m ≤? n)
suc m ≤? suc n    | right e = right (lesuc e)
suc m ≤? suc n    | left e = left (lesuc e)

{- Insertion sort -}

insert : (x : ℕ) → (l : List ℕ) → List ℕ
insert x [] = x ∷ []
insert x (hd ∷ tl) with (x ≤? hd)
insert x (hd ∷ tl)    | left e = x ∷ (hd ∷ tl)
insert x (hd ∷ tl)    | right e = hd ∷ (insert x tl)

sort : List ℕ → List ℕ
sort [] = []
sort (x ∷ l) = insert x (sort l)

test : List ℕ
test = sort (4 ∷ 1 ∷ 45 ∷ 8 ∷ 32 ∷ 12 ∷ 1 ∷ [])

data _≤*_ : (x : ℕ) → (l : List ℕ) → Set where
  ≤*-empty : (x : ℕ) → x ≤* []
  ≤*-ind :  (x hd : ℕ) → (tl : List ℕ) → x ≤ hd → x ≤* tl → x ≤* (hd ∷ tl)

data sorted : (l : List ℕ) → Set where
  sort-empty : sorted []
  sort-ind :  (hd : ℕ) → (tl : List ℕ) → hd ≤* tl → sorted tl → sorted (hd ∷ tl)

≤*-trans :  (x y : ℕ) → (l : List ℕ) → x ≤ y → y ≤* l → x ≤* l
≤*-trans x y [] = λ x₁ x₂ → ≤*-empty x
≤*-trans x y (hd ∷ tl) xley (≤*-ind .y .hd .tl ylehd ytl) = ≤*-ind x hd tl (≤-trans xley ylehd) (≤*-trans x y tl xley ytl)

≤*-insert : (x y : ℕ) → (l : List ℕ) → x ≤ y → x ≤* l → x ≤* insert y l
≤*-insert x y [] xley xlel = ≤*-ind x y [] xley xlel
≤*-insert x y (hd ∷ tl) xley xlel with (y ≤? hd)
≤*-insert x y (hd ∷ tl) xley xlel    | left ylehd = ≤*-ind x y (hd ∷ tl) xley xlel
≤*-insert x y (hd ∷ tl) xley (≤*-ind .x .hd .tl xlehd xlel) | right hdley = ≤*-ind x hd (insert y tl) xlehd (≤*-insert x y tl xley xlel)

insert-sorting : (x : ℕ) → (l : List ℕ) → sorted l → sorted (insert x l)
insert-sorting x .[] sort-empty = sort-ind x [] (≤*-empty x) sort-empty
insert-sorting x .(hd ∷ tl) (sort-ind hd tl shd stl) with (x ≤? hd)
insert-sorting x .(hd ∷ tl) (sort-ind hd tl shd stl)    | left xlehd = sort-ind x (hd ∷ tl) (≤*-ind x hd tl xlehd (≤*-trans x hd tl xlehd shd)) (sort-ind hd tl shd stl)
insert-sorting x .(hd ∷ tl) (sort-ind hd tl shd stl)    | right hdlex = sort-ind hd (insert x tl) (≤*-insert hd x tl hdlex shd) (insert-sorting x tl stl)

sort-spec :  (l : List ℕ) → sorted (sort l)
sort-spec [] = sort-empty
sort-spec (x ∷ l) = insert-sorting x (sort l) (sort-spec l)

f : List ℕ → List ℕ
f l = []

f-sorting : (l : List ℕ) → sorted (f l)
f-sorting l = sort-empty

mutual
  data Sorted : Set where
    nil : Sorted
    cons : (hd : ℕ) → (tl : Sorted) → (e : hd ≤ head hd tl)  → Sorted

  head : ℕ → Sorted → ℕ
  head d nil = d
  head d (cons hd l e) = hd

mutual
  insert' : (x : ℕ) → Sorted → Sorted
  insert' x nil = cons x nil ≤-refl
  insert' x (cons hd l e) with (x ≤? hd)
  insert' x (cons hd l e)    | left xlehd = cons x (cons hd l e) xlehd
  insert' x (cons hd l e)    | right hdlex = cons hd (insert' x l) (insert-head hd x l hdlex e)
                                            
  insert-head : (x y : ℕ) → (l : Sorted) → x ≤ y → x ≤ head x l → x ≤ head x (insert' y l)
  insert-head x y nil xley xlel = xley
  insert-head x y (cons hd l e) xley xlel with (y ≤? hd)
  ...                                        | left e' = xley
  ...                                        | right e' = xlel

sort' : (l : List ℕ) → Sorted
sort' [] = nil
sort' (x ∷ l) = insert' x (sort' l)



{- 3 : Preservation of elements -}

data _∼_ {A : Set} : List A → List A → Set where
  ∼-nil : [] ∼ []
  ∼-drop : (x : A) {l l' : List A} → l ∼ l' → (x ∷ l) ∼ (x ∷ l')
  ∼-swap : (x y : A) (l : List A) → (x ∷ y ∷ l) ∼ (y ∷ x ∷ l)
  ∼-trans : {l l' l'' : List A} → l ∼ l' → l' ∼ l'' → l ∼ l''

∼-refl : {A : Set} {l : List A} → l ∼ l
∼-refl {l = []} = ∼-nil
∼-refl {l = x ∷ l} = ∼-drop x ∼-refl

∼-sym : {A : Set} {l l' : List A} → l ∼ l' → l' ∼ l
∼-sym ∼-nil = ∼-nil
∼-sym (∼-drop x e) = ∼-drop x (∼-sym e)
∼-sym (∼-swap x y l) = ∼-swap y x l
∼-sym (∼-trans e e') = ∼-trans (∼-sym e') (∼-sym e)

insert-∼ : (x : ℕ) (l : List ℕ) → (x ∷ l) ∼ (insert x l)
insert-∼ x [] = ∼-refl
insert-∼ x (hd ∷ l) with x ≤? hd
...                    | left e = ∼-refl
...                    | right e = ∼-trans (∼-swap x hd l) (∼-drop hd (insert-∼ x l))

∼-insert : (x : ℕ) {l l' : List ℕ} → l ∼ l' → insert x l ∼ insert x l'
∼-insert x {l} {l'} e = ∼-trans (∼-sym (insert-∼ x l)) (∼-trans (∼-drop x e) (insert-∼ x l'))

sort-∼ : (l : List ℕ) → l ∼ (sort l)
sort-∼ [] = ∼-nil
sort-∼ (x ∷ l) = ∼-trans (∼-drop x (sort-∼ l)) (insert-∼ x (sort l))

{- 4 : Merge sort -}

split : {A : Set} → List A → List A × List A
split [] = [] , []
split (x ∷ []) = (x ∷ []) , []
split (x ∷ y ∷ l) = (x ∷ (fst (split l))) , (y ∷ snd (split l))

merge : (l m : List ℕ) → List ℕ
merge [] m = m
merge (x ∷ l) [] = x ∷ l
merge (x ∷ l) (y ∷ m) with x ≤? y
...                      | left xley = x ∷ (merge l (y ∷ m))
...                      | right ylex = y ∷ (merge (x ∷ l) m)

{-# TERMINATING #-}
mergesort : List ℕ → List ℕ
mergesort [] = []
mergesort (x ∷ []) = x ∷ []
mergesort (x ∷ y ∷ l) = merge (mergesort (fst (split (x ∷ y ∷ l)))) (mergesort (snd (split (x ∷ y ∷ l))))

test-merge : List ℕ
test-merge = mergesort (4 ∷ 1 ∷ 45 ∷ 8 ∷ 32 ∷ 12 ∷ 1 ∷ [])

_<_ : (m n : ℕ) → Set
m < n = suc m ≤ n


length : {A : Set} (l : List A) → ℕ
length [] = zero
length (x ∷ l) = suc (length l)

len-split-fst : (x y : ℕ) (l : List ℕ) → length (fst (split (x ∷ y ∷ l))) < suc (suc (length l))
len-split-fst x y [] = lesuc (lesuc lez)
len-split-fst x y (z ∷ []) = lesuc (lesuc (lesuc lez))
len-split-fst x y (z ∷ w ∷ l) = ≤-suc (≤-trans (len-split-fst z w l) (≤-suc (≤-suc (≤s (length l)))))

len-split-snd : (x : ℕ) (l : List ℕ) → length (snd (split (x ∷ l))) < suc (length l)
len-split-snd x [] = lesuc lez
len-split-snd x (z ∷ []) = lesuc (lesuc lez)
len-split-snd x (y ∷ z ∷ l) = ≤-suc (≤-trans (len-split-snd z l) (≤-suc (≤s (length l)) ))

mergesort-fuel : (n : ℕ) → (l : List ℕ) → (length l ≤ n) → List ℕ
mergesort-fuel n [] e = []
mergesort-fuel n (x ∷ []) e = x ∷ []
mergesort-fuel (suc n) (x ∷ y ∷ l) e = merge (mergesort-fuel n (fst (split (x ∷ y ∷ l))) (≤-pred (≤-trans (len-split-fst x y l) e)) )
                                       (mergesort-fuel n (snd (split (x ∷ y ∷ l)))(≤-pred (≤-trans (len-split-snd x (y ∷ l)) e)))

new-mergesort : (l : List ℕ) → List ℕ
new-mergesort l = mergesort-fuel (length l) l ≤-refl

≤*-merge :  (l m : List ℕ) → (x : ℕ) → x ≤* l → x ≤* m → x ≤* merge l m
≤*-merge .[] m x (≤*-empty .x) sm = sm
≤*-merge .(hd ∷ tl) .[] x (≤*-ind .x hd tl sx sl) (≤*-empty .x) = ≤*-ind x hd tl sx sl
≤*-merge .(hd ∷ tl) .(hd' ∷ tl') x (≤*-ind .x hd tl sx sl) (≤*-ind .x hd' tl' sx' sm) with hd ≤? hd'
...                          | left e = ≤*-ind x hd (merge tl (hd' ∷ tl')) sx (≤*-merge tl (hd' ∷ tl') x sl (≤*-ind x hd' tl' sx' sm))
...                          | right e = ≤*-ind x hd' (merge (hd ∷ tl) tl') sx' (≤*-merge (hd ∷ tl) tl' x (≤*-ind x hd tl sx sl) sm)

merge-spec : (l m : List ℕ) → sorted l → sorted m → sorted ( merge l m)
merge-spec [] m sl sm = sm
merge-spec (x ∷ l) [] sl sm = sl
merge-spec (x ∷ l) (y ∷ m) sl sm  with x ≤? y
merge-spec (x ∷ l) (y ∷ m) (sort-ind .x .l xle sl) (sort-ind .y .m yle sm) | left e = sort-ind x (merge l (y ∷ m))
                                                                                      (≤*-merge l (y ∷ m) x xle
                                                                                      (≤*-ind x y m e (≤*-trans x y m e yle)))
                                                                                      (merge-spec l (y ∷ m) sl (sort-ind y m yle sm))
merge-spec (x ∷ l) (y ∷ m) (sort-ind .x .l xle sl) (sort-ind .y .m yle sm) | right e = sort-ind y (merge (x ∷ l) m)
                                                                                       (≤*-merge (x ∷ l) m y (≤*-trans y x (x ∷ l) e (≤*-ind x x l ≤-refl xle)) yle)
                                                                                       (merge-spec (x ∷ l) m (sort-ind x l xle sl) sm)


mergesort-fuel-spec : (n : ℕ) → (l : List ℕ) → (e : length l ≤ n) →  sorted (mergesort-fuel n l e)
mergesort-fuel-spec n [] e = sort-empty
mergesort-fuel-spec n (x ∷ []) e = sort-ind x [] (≤*-empty x) sort-empty
mergesort-fuel-spec (suc n) (x ∷ y ∷ l) e = merge-spec (mergesort-fuel n (fst (split (x ∷ y ∷ l))) (≤-pred (≤-trans (len-split-fst x y l) e)))
                                            (mergesort-fuel n (snd (split (x ∷ y ∷ l)))  (≤-pred (≤-trans (len-split-snd x (y ∷ l)) e)))
                                            (mergesort-fuel-spec n (x ∷ fst (split l))  (≤-pred (≤-trans (len-split-fst x y l) e)))
                                            (mergesort-fuel-spec n ((y ∷ snd (split l))) (≤-pred (≤-trans (len-split-snd x (y ∷ l)) e)))



mergesort-sorting : (l : List ℕ) → sorted (new-mergesort l)
mergesort-sorting [] = sort-empty
mergesort-sorting (x ∷ []) = sort-ind x [] (≤*-empty x) sort-empty
mergesort-sorting (x ∷ y ∷ l) = merge-spec
                                (mergesort-fuel (suc (length l)) (x ∷ fst (split l)) (≤-pred (≤-trans (len-split-fst x y l) ≤-refl)))
                                (mergesort-fuel (suc (length l)) (snd (split (x ∷ y ∷ l))) (≤-pred (≤-trans (len-split-snd x (y ∷ l)) ≤-refl)))
                                (mergesort-fuel-spec (suc (length l)) ((x ∷ fst (split l))) (≤-pred (≤-trans (len-split-fst x y l) ≤-refl)) )
                                (mergesort-fuel-spec (suc (length l)) (y ∷ snd (split l)) (≤-pred (≤-trans (len-split-snd x (y ∷ l)) ≤-refl)))

{- 5 : Well-founded definition of merge sort -}

Rel : (A : Set) → Set₁
Rel A = A → A → Set

data Acc {A : Set} (_<_ : Rel A) (x : A) : Set where
  acc : ((y : A) → y < x → Acc _<_ y) → Acc _<_ x

WellFounded : {A : Set} → (_<_ : Rel A) → Set
WellFounded {A} _<_ = (x : A) → Acc _<_ x

≤-< : {m n : ℕ} → (m ≤ n) → m ≡ n ∨ m < n
≤-< {n = zero} lez = left refl
≤-< {n = suc n} lez = right (lesuc lez)
≤-< (lesuc {m} {n} e) with ≤-< e
...                      | left x = left (cong suc x)
...                      | right x = right (≤-suc x)



<-last : {m n : ℕ} → (m < suc n) → m ≡ n ∨ m < n
<-last (lesuc e) = ≤-< e

apply : {A B : Set} → (A → B) → A → B
apply = λ x → x

rec : (P : ℕ → Set) → P zero → (Pn : (n : ℕ) → P n → P (suc n)) → (n : ℕ) → P n
rec P Pzero Pn zero = Pzero
rec P Pzero Pn (suc n) = Pn n (rec P Pzero Pn n)

{-
strong-ind : (P : ℕ → Set) → P zero → (Pn : (n : ℕ) → ((m : ℕ) → m ≤ n → P m) →  ((m : ℕ) → m ≤ suc n → P m) ) → (n m : ℕ) → m ≤ n → P m
strong-ind P Pzero Pn n = apply (rec {! (m : ℕ) → (m ≤ n → P m)!} {!!} {!!}) n ≤-refl-}
{-rec (λ x → {!(m : ℕ) → m ≤ x → P m!}) {!!} {!!} {!!}-}

wfNat-aux : (n y : ℕ) → Acc _<_ n → y < suc n → Acc _<_ y
wfNat-aux n y accn yltn with <-last yltn
wfNat-aux n .n accn yltn | left refl = accn
wfNat-aux n y (acc accn) yltn | right e = accn y e

wfNat : WellFounded _<_
wfNat = rec (λ x → Acc _<_ x) (acc (λ y ())) λ n accn → acc λ y ylt → wfNat-aux n y accn ylt


mergesort-WF-aux : (l : List ℕ) → Acc _<_ (length l) → List ℕ
mergesort-WF-aux [] accl = []
mergesort-WF-aux (x ∷ []) accl = x ∷ []
mergesort-WF-aux (x ∷ y ∷ l) (acc accl) = merge (mergesort-WF-aux (fst (split (x ∷ y ∷ l))) (accl (length (x ∷ fst (split l))) (len-split-fst x y l)))
                                            (mergesort-WF-aux (snd (split (x ∷ y ∷ l))) (accl (length (snd (split (x ∷ y ∷ l)))) (len-split-snd y (y ∷ l))))

mergesort-WF :  (l : List ℕ) → List ℕ
mergesort-WF l = mergesort-WF-aux l (wfNat (length l))


